package org.herac.tuxguitar.player.impl.sequencer;

import org.herac.tuxguitar.player.base.MidiPlayerException;
import org.herac.tuxguitar.player.base.MidiPort;
import org.herac.tuxguitar.player.base.MidiPortEmpty;
import org.herac.tuxguitar.player.base.MidiSequenceHandler;
import org.herac.tuxguitar.player.base.MidiSequencer;
import org.herac.tuxguitar.util.TGLock;

public class MidiSequencerImpl implements MidiSequencer{

	private static final TGLock lock = new TGLock();

	private boolean reset;
	private boolean running;
	private MidiPort midiPort;
	private MidiTickPlayer midiTickPlayer;
	private MidiEventPlayer midiEventPlayer;
	private MidiEventDispacher midiEventDispacher;
	private MidiTrackController midiTrackController;
	
	public MidiSequencerImpl(){
		this.running = false;
		this.midiTickPlayer = new MidiTickPlayer();
		this.midiEventPlayer = new MidiEventPlayer(this);
		this.midiEventDispacher = new MidiEventDispacher(this);
		this.midiTrackController = new MidiTrackController(this);
	}
	
	public MidiTrackController getMidiTrackController(){
		return this.midiTrackController;
	}
	
	public void setTempo(int tempo){
		this.midiTickPlayer.setTempo(tempo);
	}
	
	public long getTickPosition(){
		return this.midiTickPlayer.getTick();
	}		
	
	public long getTickLength(){
		return this.midiTickPlayer.getTickLength();
	}
	
	public void setTickPosition(long tickPosition){
		this.reset = true;
		this.midiTickPlayer.setTick(tickPosition);
	}

	public void sendEvent(MidiEvent event) throws MidiPlayerException{
		if(!this.reset){
			this.midiEventDispacher.dispatch(event);
		}
	}
	
	public void addEvent(MidiEvent event){
		this.midiEventPlayer.addEvent(event);
		this.midiTickPlayer.notifyTick(event.getTick());
	}	
	
	public boolean isRunning() {
		return this.running;
	}

	public synchronized void setRunning(boolean running) throws MidiPlayerException {
		this.running = running;
		this.setTempo(120);
		this.setTickPosition(0);
		if(this.running){
			new MidiTimer(this).start();
		}
		else{
			this.midiEventPlayer.clearEvents();
			this.midiTickPlayer.clearTick();
			this.getMidiPort().out().sendAllNotesOff();
			this.getMidiPort().out().sendSystemReset();
		}
	}
	
	public synchronized void stop() throws MidiPlayerException{
		this.setRunning(false);
	}
	
	public synchronized void start() throws MidiPlayerException{
		this.setRunning(true);
	}

	protected synchronized boolean process() throws MidiPlayerException{
		boolean running = this.isRunning();
		if(running && !lock.isLocked()){
			lock.lock();

			if(this.reset){
				this.reset = false;
				this.getMidiPort().out().sendAllNotesOff();
				for(int channel = 0; channel < 16;channel ++){
					this.getMidiPort().out().sendPitchBend(channel, 64);
				}
				this.midiEventPlayer.reset();
			}
				
			this.midiEventPlayer.process();
			this.midiTickPlayer.process();		
			if(this.getTickPosition() > this.getTickLength()){
				this.stop();
			}

			lock.unlock();
		}
		return running;
	}

	public MidiPort getMidiPort() {
		if(this.midiPort == null){
			this.midiPort = new MidiPortEmpty();
		}
		return this.midiPort;
	}

	public void setMidiPort(MidiPort midiPort) {
		this.midiPort = midiPort;
	}

	public void open() {
		//not implemented		
	}

	public void close() throws MidiPlayerException {
		if(isRunning()){
			this.stop();
		}
	}
	
	public MidiSequenceHandler createSequence(int tracks){
		return new MidiSequenceHandlerImpl(this,tracks);
	}
	
	public void setSolo(int index,boolean solo) throws MidiPlayerException{
		this.getMidiTrackController().setSolo(index, solo);
	}
	
	public void setMute(int index,boolean mute) throws MidiPlayerException{
		this.getMidiTrackController().setMute(index, mute);
	}

	public String getKey() {
		return "tuxguitar.sequencer";
	}

	public String getName() {
		return "TuxGuitar Sequencer";
	}

	private class MidiTimer extends Thread{

		private static final int TIMER_DELAY = 15;

		private MidiSequencerImpl sequencer;
		
		public MidiTimer(MidiSequencerImpl sequencer){
			this.sequencer = sequencer;
		}

		public void run() {
			try {
				synchronized(this.sequencer) {
					while( this.sequencer.process() ){
						this.sequencer.wait( TIMER_DELAY );
					}
				}
			} catch (Throwable throwable) {
				throwable.printStackTrace();
			}
		}
	}
}
